/**
* CMU MSUAV Project (Micro and Small Unmanned Aerial Vehicle)
*          
* \file  featurePool.h
* \brief implementation of CFeature2D, CFeature2DTrack, and CFeature2DPool
* \date  15-Jun-2007
*
* \author Myung Hwangbo (myung@cs.cmu.edu) and Jun-Sik Kim (kimjs@cs.cmu.edu)
*
* Copyright (c) 2007-2009 Myung Hwangbo and Jun-Sik Kim
* Robotics Institute, Carnegie Mellon University
*
*/

#include <ipp.h>
#include <algorithm>
#include "featurePool.h"

#include <cuda_runtime.h>

#include <boost/interprocess/managed_shared_memory.hpp>
#include <boost/interprocess/sync/interprocess_barrier.hpp>
#include <boost/interprocess/sync/interprocess_mutex.hpp>



#include "cuda_feature_select.h"

#include "ctimer.h"

#include "litmus.h"
#include "chunk.h"

#if 0
#define CUDA_SAFE_CALL(call) {                                         \
	cudaError err = call;                                              \
	if (cudaSuccess != err) {                                          \
		fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n",  \
			__FILE__, __LINE__, cudaGetErrorString( err) );            \
		exit(EXIT_FAILURE);                                            \
    }} 
#else
#define CUDA_SAFE_CALL(call) { call; }
#endif

//extern "C" int InitCUDA(int imageWidth, int imageHeight, int pyrLevel, int templateSize, int maxIteration, int maxLinesearch, int numGPUs, int gpuOffset, int numUsers);
extern "C" int InitCUDA(int imageWidth, int imageHeight, int pyrLevel, int templateSize, int maxIteration, int maxLinesearch, int numGPUs, int gpuPartition, int numGPUPartitions, int numUsers);
extern "C" bool ExitCUDA(void);
extern "C" void Caller_Tracker_Registration(float* fx, float *fy, int numfeatures, int *indexmap);
extern "C" void Caller_Align(cudaArray *Image2, float* motion, float *residual, float *vec_invH);
extern "C" bool MigrateToGPU_Implicit(int to);
extern "C" void set_cur_gpu(int gpu);

#define MAX_PYR_LEVEL		5
extern cudaArray *Image1[NR_GPUS][MAX_PYR_LEVEL], *Image2[NR_GPUS][MAX_PYR_LEVEL];
extern void *Image1Temp[NR_GPUS][MAX_PYR_LEVEL], *Image2Temp[NR_GPUS][MAX_PYR_LEVEL];
extern size_t Image1TempPitch[NR_GPUS][MAX_PYR_LEVEL], Image2TempPitch[NR_GPUS][MAX_PYR_LEVEL];

extern cudaStream_t STREAMS[NR_GPUS];
extern int CUR_DEVICE;

extern int GPU_PARTITION;
extern int GPU_PARTITION_SIZE;
extern int NUM_SIMULT_USERS;

extern int EE_LOCKS[NR_GPUS];
extern int CE_SEND_LOCKS[NR_GPUS];
extern int CE_RECV_LOCKS[NR_GPUS];

extern "C" void allocate_locks(void);

extern int KEXCLU_LOCK;

static inline bool useEngineLocks(void)
{
	return(NUM_SIMULT_USERS != 1);
}


#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>

static void* allocStart;
static size_t allocSize;








//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

using namespace boost::interprocess;

gpu_pool* gpu_pool::gpuLinuxSemPools = NULL;

void gpu_pool::initLinuxSemPools(int numPools, int poolSize, bool create)
{
	managed_shared_memory *segment_ptr;

	if(create)
	{
		printf("%d creating shared memory for linux semaphores\n", getpid());
		shared_memory_object::remove("linux_sem_memory");
		segment_ptr = new managed_shared_memory(create_only, "linux_sem_memory", 4*1024);
		
		gpuLinuxSemPools = segment_ptr->construct<gpu_pool>("gpu_pool p")[numPools](poolSize);
	}
	else
	{
		do
		{
			try
			{
				printf("%d opening shared memory for linux semaphores\n", getpid());
				segment_ptr = new managed_shared_memory(open_only, "linux_sem_memory");
			}
			catch(...)
			{
				printf("%d shared memory not ready.  sleeping\n", getpid());
				sleep(1);
			}
		}while(segment_ptr == NULL);
		
		gpuLinuxSemPools = segment_ptr->find<gpu_pool>("gpu_pool p").first;
	}		
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

CFeature2DPool* CFeature2DPool::CreateCFeature2DPool(CvSize imgSize, int templSize, int pyrLevel, int maskSize, bool gpu, CCameraModel* pCameraModel,
													 int numGPUs, int gpuPartition, int numGPUPartitions, int numUsers, int lockMode)
{
	size_t pages = sizeof(CFeature2DPool) >> 12;
	if(sizeof(CFeature2DPool)%(1<<12) != 0)
		pages += 1;  // round up

	allocSize = pages*(1<<12);
	allocStart = mmap(NULL, allocSize, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS | MAP_LOCKED, -1, 0);
	
	CFeature2DPool* instance = new (allocStart) CFeature2DPool(imgSize, templSize, pyrLevel, maskSize, gpu, pCameraModel,
															   numGPUs, gpuPartition, numGPUPartitions, numUsers, LITMUS_SEM);

	//printf("FeatureBuff @ %p / %p : image size = %d, reg size = %d\n", allocStart, instance, (int)sizeof(CFeature2DPool), (int)allocSize);

	if (gpu) {
		cudaError_t ret = cudaHostRegister(allocStart, allocSize, cudaHostRegisterPortable);
	}

	//printf("registration of CFeature2DPool memory: %d\n", ret);

	return(instance);
}

void CFeature2DPool::DestroyCFeature2DPool(CFeature2DPool** pool)
{
	// explicitly call the destructor
	if(pool && *pool)
	{
		(*pool)->~CFeature2DPool();
		*pool = NULL;
		
		// TODO: Free up the mmap() buffer, inform CUDA that the memory is gone??
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
//!
//! CFeature2DPool class
//!
CFeature2DPool::CFeature2DPool(CvSize imgSize, int templSize, int pyrLevel, int maskSize, bool gpu, CCameraModel* pCameraModel,
							   int numGPUs, int gpuPartition, int numGPUPartitions, int numUsers, int lockMode)
	: m_imgSize(imgSize), m_templateSize(templSize), m_pyrLevel(pyrLevel), m_maskSize(maskSize),
	  b_gpu(gpu), m_pCameraModel(pCameraModel),
	  m_kexcluLock(-1), m_locked(false), m_gpuOffset(gpuPartition*numGPUs), m_curGPU(gpuPartition*numGPUs), m_lockMode((eLockMode)LITMUS_SEM)
{
	reserve(50000);

	m_frame = 0;
	m_pyr = NULL;

	// Memory alloc for OPENCV feature selection, see cvGoodFeaturesToTrack()
	m_imgEig  = cvCreateImage(m_imgSize, IPL_DEPTH_32F, 1);
	m_imgTmp  = cvCreateImage(m_imgSize, IPL_DEPTH_32F, 1);
	m_imgMask = cvCreateImage(m_imgSize, IPL_DEPTH_8U, 1);

	if (b_gpu)
	{
		m_pyrLevel = InitCUDA(m_imgSize.width, m_imgSize.height, m_pyrLevel, templSize, 3, 5, numGPUs, gpuPartition, numGPUPartitions, numUsers);		
		
		allocate_locks();
		m_kexcluLock = KEXCLU_LOCK;
				
		for (int i=0;i < MAX_NUM_FEATURES;i++) {
			m_cuda_index_map_global[i] = -1;	// index mapping from cuda mem to global mem
			m_cuda_index_map_lost[i] = 1;		// index of lost features of cuda mem (lost 1, tracked 0)
			m_cuda_index_map[i] = 0;			// index map from new features to cuda mem
		}


		// HACK.  Everything is page-locked in Litmus.  This class is on the heap.
		// So let's register our memory with CUDA as being pinned.
		// TODO: CUDA 4.1: Registrations do not need to be page aligned.

		uint64_t pagesStart_4k, pagesEnd_4k;
		size_t bytes;
		cudaError_t ret;

		IplImage* toreg[3] = {m_imgEig, m_imgTmp, m_imgMask};
		for(int i = 0; i < sizeof(toreg)/sizeof(toreg[0]); ++i)
		{
			IplImage* reg = toreg[i];
			// register image data alloc'ed by OpenCV
			pagesStart_4k = ((uint64_t)reg->imageData)>>12;
			pagesEnd_4k = ((uint64_t)&(reg->imageData[reg->imageSize]))>>12;
			if(((uint64_t)&(reg->imageData[reg->imageSize]))%(1<<12) != 0)
				pagesEnd_4k += 1;  // round up to next page
			bytes = (pagesEnd_4k - pagesStart_4k)<<12;
			//printf("reg image data @ %p : image size = %d, reg size = %d\n", reg->imageData, reg->imageSize, (int)bytes);
			ret = cudaHostRegister((void*)(pagesStart_4k<<12), bytes, cudaHostRegisterPortable);
			//printf("reg return code = %d\n", ret);
		}

	}
}

void CFeature2DPool::cleanUpLocks()
{	
	if (b_gpu) {
//		if (m_lockMode == LINUX_SEM) {
//			m_locked = false;
//			m_linuxSemPool->put(m_curGPU - m_gpuOffset);
//		}

		if (m_lockMode == LITMUS_SEM) {
			// try to unlock everything.  litmus will prevent bogus calls.
			if(NUM_SIMULT_USERS > 1)
			{
				for(int i = 0; i < GPU_PARTITION_SIZE; ++i)
				{
					int which_device = GPU_PARTITION*GPU_PARTITION_SIZE + i;
					
					litmus_unlock(EE_LOCKS[which_device]);
					litmus_unlock(CE_SEND_LOCKS[which_device]);
					if(NUM_SIMULT_USERS == 2) {
						litmus_unlock(CE_RECV_LOCKS[which_device]);
					}
				}
			}
			
			if(CUR_DEVICE >= 0) {
				unregister_nv_device(CUR_DEVICE);
			}
			
			litmus_unlock(KEXCLU_LOCK);
		}

	}
}

CFeature2DPool::~CFeature2DPool()
{
    if (b_gpu) {
		this->cleanUpLocks();
	}

	/*

	for (int i=0; i < (int)(this->size()); i++) delete this->at(i);
	
	map<int, IplImage*>::iterator im = m_mapImage.begin();
	for ( ; im != m_mapImage.end(); ++im) {
		cvReleaseImage(&im->second);
	}

	// Memory release for OPENCV feature selection, see cvGoodFeaturesToTrack()
	cvReleaseImage(&m_imgEig);
	cvReleaseImage(&m_imgTmp);
	cvReleaseImage(&m_imgMask);

	*/
}

int CFeature2DPool::AddImage(int frame, IplImage* img)
{
	m_frame = frame;
	m_img = img;
	m_mapImage[frame] = img;

	// Create a new pyramid image
	if (m_pyr) delete m_pyr;
	m_pyr = new CImagePyramid(img, m_pyrLevel);

	return (int)m_mapImage.size();
}

int CFeature2DPool::PurgeImage(int frame, int lag_dist)
{
	map<int, IplImage*>::iterator im;

	for (im = m_mapImage.begin(); im != m_mapImage.end(); im++) {
		if (im->first < frame - lag_dist) {
			cvReleaseImage(&im->second);
			m_mapImage.erase(im++);
		}
	}

	return (int)m_mapImage.size();
}

const IplImage* CFeature2DPool::GetImage(int frame)
{
	if (frame < 0) return m_mapImage[m_frame]; // last image
	else {
		map<int, IplImage*>::const_iterator im = m_mapImage.find(frame);
		if (im == m_mapImage.end()) return NULL;
		else return im->second;
	}
}

CFeature2DTrack* CFeature2DPool::GetFeatureTrackPtr(int id)
{
	if (id < 0 || id >= (int)size()) return NULL;
	return (*this)[id];
}

CFeature2DTrack& CFeature2DPool::GetFeatureTrack(int id)
{
	assert(id >= 0 && id < (int)size());
	return *(*this)[id];
}

CFeature2D* CFeature2DPool::GetFeaturePtr(int id, int frame)
{
	if (id < 0 || id >= (int)size()) return NULL;
	return (*this)[id]->GetFeaturePtr(frame);
}

CFeature2D& CFeature2DPool::GetFeature(int id, int frame)
{
	assert(id >= 0 && id < (int)size());
	return (*this)[id]->GetFeature(frame);
}

void CFeature2DPool::SetOutlier(int id, int inout)
{
	m_mapOutliers[id] = inout;
}

int CFeature2DPool::GetOutlier(vector<int>& vid)
{
	vid.clear();
	map<int, int>::const_iterator im;
	for (im = m_mapOutliers.begin(); im != m_mapOutliers.end(); im++) {
		if (im->second == 1) {
			vid.push_back(im->first);
		}
	}

	return (int)vid.size();
}

bool CFeature2DPool::IsOutlier(int id)
{
	map<int, int>::const_iterator im = m_mapOutliers.find(id);
	if (im == m_mapOutliers.end()) {
		return false;
	}
	else {
		return (im->second == 1);
	}
}

void CFeature2DPool::ComputeSelectionMask(int templ_size, int mask_size, const list<int>& activeID, IplImage* img_mask)
{
    //if (!img_mask) return;
    
	int width  = img_mask->width;
	int height = img_mask->height;

	int hw = templ_size >> 1;

	//int x_border = (hw+1) << (MAX_PYRAMID_LEVEL - 1);
	//int y_border = (hw+1) << (MAX_PYRAMID_LEVEL - 1);
	int x_border = (hw+1) << (m_pyrLevel - 1);
	int y_border = (hw+1) << (m_pyrLevel - 1);

	int wx = mask_size;
	int wy = mask_size;

	int bottom_offset = 0;

	cvSetZero(img_mask);
	cvSetImageROI(img_mask, cvRect(x_border, y_border, m_imgSize.width - 2*x_border, m_imgSize.height - 2*y_border - bottom_offset));
	cvSet(img_mask, cvScalar(255));

	// Generate the mask from current feature points
	//cout << "size = " << m_listActiveID.size() << endl;
	for (list<int>::const_iterator it(activeID.begin()), theEnd(activeID.end()); it != theEnd; ++it) 
	{
		int id = *it;
		CFeature2D* fp = GetFeaturePtr(id);	//cout << id << endl;

		CvPoint p1, p2;
		p1.x = int(fp->m_x - wx/2 + 0.5);	p1.y = int(fp->m_y - wy/2 + 0.5);
		p2.x = int(fp->m_x + wx/2 + 0.5);	p2.y = int(fp->m_y + wy/2 + 0.5);

		p1.x = max(0, min(p1.x, width -1));
		p1.y = max(0, min(p1.y, height-1));
		p2.x = max(0, min(p2.x, width -1));
		p2.y = max(0, min(p2.y, height-1));
		if (p1.x == p2.x || p1.y == p2.y) continue;

		cvSetImageROI(img_mask, cvRect(p1.x, p1.y, p2.x-p1.x, p2.y-p1.y));
		cvSetZero(img_mask);
	}

	// Don't forget to reset the ROI of imgFeatureMask before running cvGoodFeaturesToTrack()
	cvResetImageROI(img_mask);
}

int CFeature2DPool::Select(int frame, int min_threshold, int max_num, IplImage* gimgf, bool verbose)
{
	assert(min_threshold <= max_num && frame >= 0);

	int f_num = (int)m_listActiveID.size();
	if (f_num >= min_threshold) return 0;
	
	int n_new = max_num - f_num;
	int block_size = 3;
	double min_distance = 20;
	double quality_level = 0.005;

	//! Select new features in the area where no other features exist nearby.
	CvPoint2D32f* corners = new CvPoint2D32f[n_new];
	ComputeSelectionMask(m_templateSize, m_maskSize, m_listActiveID, m_imgMask); 
	cvGoodFeaturesToTrack(gimgf, m_imgEig, m_imgTmp, corners, &n_new, quality_level, min_distance, m_imgMask, block_size);

	//! Add new features to the feature pool
	for (int i=0; i < n_new; i++) {
		float saliency = CV_IMAGE_ELEM(m_imgEig, float, (int)corners[i].y, (int)corners[i].x);
		int nid = Add(corners[i].x, corners[i].y, saliency, frame, m_pyr);
	}
	delete [] corners;

	if (0) {
		cout << "CFeature2DPool::Select() : frame = " << frame 
			<< ", f_num = " << f_num << ", new_num = " << n_new << ", total_num = " << (int)m_listActiveID.size() << endl;
	}

	m_stat.time_select = 0;
	m_stat.n_new = n_new;
	m_mapStat[frame] = m_stat;

	//! Update activeID map so that newly selected features are included.
	m_mapActiveID[frame] = m_listActiveID;

	return n_new;
}

int CFeature2DPool::Add(float x, float y, float saliency, int frame, CImagePyramid* newPyr)
{
	//! 1. A unique ID is assigned to a new feature tracker
	int new_id = (int)size();

	//! 2. Create a new feature tracker
	CFeature2DTrack *pTrack = new CFeature2DTrack(x, y, saliency, frame, new_id, m_templateSize, newPyr, m_pCameraModel);

	//! 3. Append to the feature pool
	push_back(pTrack);

	//! 4. Append the new ID to the active feature list
	m_listActiveID.push_back(new_id);

	return new_id;
}

int CFeature2DPool::Purge(int frame, int lag_dist)
{
	PurgeImage(frame, lag_dist);

	//! Note that it purges features that are lost at least "lag_dist" ago.
	//! Be cautious that the purged features are not completely deleted from the feature pool
	//! because erasing elements in std::vector container is expensive.
	//! Instead the memories that they have used are deallocated.

	int num_purged = 0;
	
	list<int>::iterator it = m_listLostID.begin();
	while (it != m_listLostID.end()) 
	{
		int id = (int)(*it);
		int d = frame - (*this)[id]->GetEndFrame();
		if (d > lag_dist) {
			(*this)[id]->~CFeature2DTrack();
			(*this)[id] = NULL;
			it = m_listLostID.erase(it);

			++num_purged;
		}
		else {
			++it;
		}
	}

	return num_purged;
}




//extern int //chpt;

void CFeature2DPool::Track(int frame, int timestamp, IplImage* gimgf, double* Hmg)
{
	//! Tracking, updating template, and refreshing active ID list happen at the same time.
	//! This is a single thread version that aligns all the active features sequentially.

//	//chpt = 8000;

	//! Predictive Homography
	CvMat* H = NULL;
	if (Hmg != NULL) {
		H = cvCreateMat(4, 4, CV_32F);
		cvSetIdentity(H);
		for (int i=0; i < 3; i++) {
			for (int j=0; j < 3; j++) {
				cvmSet(H, (i==2) ? 3 : i, (j==2) ? 3 : j, Hmg[i + 3*j]/Hmg[8]);
			}
		}
	}

//	//chpt = 8010;

	//! Copy feature ID
	m_listLastID = m_listActiveID;

	list<int>::iterator it = m_listActiveID.begin();
	int n_total = (int)m_listActiveID.size();
	int n_success = 0;

	while (it != m_listActiveID.end())
	{
//		//chpt = 8020;

		CFeature2DTrack* pTracker = GetFeatureTrackPtr((int)(*it));

//		//chpt = 8025;
		
		// A new feature
		if (pTracker->GetStartFrame() == frame) continue;

//		//chpt = 8030;

		// Run KLT tracker
		int status = pTracker->Run(frame, m_pyr, H);

//		//chpt = 8040;

		// Failure in tracking
		if (status & CImageAlign::TRACK_FAIL) {
//			//chpt = 8050;

			m_listLostID.push_back((int)(*it));
			it = m_listActiveID.erase(it);

//			//chpt = 8060;
		}
		else {
			if (status & CImageAlign::TRACK_SUCCESS) {
//				//chpt = 8070;
				if (status == CImageAlign::TRACK_SUCCESS_TEMPL_UPDATE) {
					pTracker->UpdateTemplate();
//					//chpt = 8080;
				}
				n_success++;
			}
			else {
				cout << "Error, invalid status = " << status << endl;
			}
//			//chpt = 8090;
			++it;
		}
	}
	
	m_stat.time_select = 0;
	m_stat.n_new = 0;
	m_stat.frame = frame;
	m_stat.n_success = n_success;
	m_stat.n_total = n_total;
	m_stat.time_track = 0;
		
	m_mapStat[frame] = m_stat;
	m_mapActiveID[frame] = m_listActiveID;

//	//chpt = 8100;

	if (H != NULL) {
		cvReleaseMat(&H);

//		//chpt = 8110;
	}

//	//chpt = 8120;
}

list<int>& CFeature2DPool::GetLastIDs()
{
	return m_listLastID;
}

list<int>& CFeature2DPool::GetActiveIDs(int frame)
{
	if (frame == -1) return m_listActiveID;
	else {
		return m_mapActiveID[frame];
	}
}

CFeatureTrackStat& CFeature2DPool::GetStat(int frame)
{
	if (frame == -1) return m_stat;
	else {
		return m_mapStat[frame];
	}
}

void CFeature2DPool::SaveToFile(int timestamp, const char* fname)
{
	static ofstream fsave(fname);
	
	CFeatureTrackStat stat = GetStat();

	fsave << stat.frame << "  " << timestamp << endl;
	fsave << stat.n_new << "  " << stat.n_success << "  " << stat.n_total << endl;
	fsave << setprecision(2) << setiosflags(ios::fixed) << stat.time_select*1e3 << "  ";
	fsave << setprecision(2) << setiosflags(ios::fixed) << stat.time_track*1e3 << endl;
	
	list<int> listID = GetActiveIDs();
	fsave << listID.size() << endl;
	
	for (list<int>::iterator it = listID.begin(); it != listID.end(); it++)
	{
		int id = *it;
		CFeature2DTrack* pTrack = GetFeatureTrackPtr(id);
		if (pTrack == NULL) continue;

		int n = (int)pTrack->size();
		if (n < 1) continue;

		CFeature2D* fp = pTrack->back();
		fsave << id << "\t";
		fsave << setprecision(2) << setiosflags(ios::fixed) << setw(12) << fp->m_x;
		fsave << setprecision(2) << setiosflags(ios::fixed) << setw(12) << fp->m_y;
		fsave << setprecision(5) << setiosflags(ios::fixed) << setw(12) << fp->m_nx;
		fsave << setprecision(5) << setiosflags(ios::fixed) << setw(12) << fp->m_ny;
		fsave << endl;
	}
}

int CFeature2DPool::Select_GPU(int frame, int min_threshold, int max_num, IplImage* img, bool verbose)
{
//	assert(min_threshold <= max_num && frame >= 0);

	int f_num = (int)m_listActiveID.size();

	if (f_num >= min_threshold) {
		m_locked = false;
		litmus_unlock(m_kexcluLock);
		return 0;
	}
	
	int n_new = max_num - f_num;
	int block_size = 3;
	double min_distance = 20;
	double quality_level = 0.005;
	
	// Select new features in the area where no current feature exists.
	ComputeSelectionMask(m_templateSize, m_maskSize, m_listActiveID, m_imgMask);
	
#if 1
	CUDA_SAFE_CALL( chunkMemcpy2D(Image1Temp[CUR_DEVICE][0], Image1TempPitch[CUR_DEVICE][0],
								  img->imageData, img->width*sizeof(float), //0,
								  img->width*sizeof(float), img->height,
								  cudaMemcpyHostToDevice,
								  CUR_DEVICE,
								  useEngineLocks()));
	
	// Copy the data to array memory.  Local copies on the GPU seem to only
	// invoke the copy engine, not any kernel.  We don't to chunking here.
	// However, the copy should (?) be fast.
	if(useEngineLocks()) litmus_lock(CE_SEND_LOCKS[CUR_DEVICE]);
	CUDA_SAFE_CALL( cudaMemcpy2DToArrayAsync(Image1[CUR_DEVICE][0], 0, 0, 
											 Image1Temp[CUR_DEVICE][0], Image1TempPitch[CUR_DEVICE][0],
											 img->width*sizeof(float), img->height,
											 cudaMemcpyDeviceToDevice, STREAMS[CUR_DEVICE]));
	CUDA_SAFE_CALL( cudaStreamSynchronize(STREAMS[CUR_DEVICE]) );
	if(useEngineLocks()) litmus_unlock(CE_SEND_LOCKS[CUR_DEVICE]);
#else
	CUDA_SAFE_CALL( cudaMemcpy2DToArrayAsync(Image1[CUR_DEVICE][0], 0, 0,
											 img->imageData, img->width*sizeof(float), 
											 img->width*sizeof(float), img->height,
											 cudaMemcpyHostToDevice, STREAMS[CUR_DEVICE]));	
#endif
	
//	
//	CUDA_SAFE_CALL( cudaMemcpy2DToArrayAsync(Image1[CUR_DEVICE][0], 0, 0, img->imageData, img->width*sizeof(float), 
//										img->width*sizeof(float), img->height, cudaMemcpyHostToDevice, STREAMS[CUR_DEVICE]));

	cvGoodFeaturesToTrack_GPU(img, m_imgEig, m_imgTmp, m_corners, &n_new, quality_level, min_distance, m_imgMask, block_size);
//	cvGoodFeaturesToTrack(img, m_imgEig, m_imgTmp, corners, &n_new, quality_level, min_distance, m_imgMask, block_size);
	
	int free_index = 0;
	for (int i = 0; i < n_new; ++i) 
	{
		float saliency = CV_IMAGE_ELEM(m_imgEig, float, (int)m_corners[i].y, (int)m_corners[i].x);
		int nid = Add(m_corners[i].x, m_corners[i].y, saliency, frame, NULL);
		while ((m_cuda_index_map_lost[free_index] == 0) && (free_index < MAX_NUM_FEATURES)) ++free_index;

		m_cuda_index_map[i]= free_index;
		m_cuda_index_map_global[free_index] = nid;
		m_cuda_index_map_lost[free_index] = 0;
		m_cuda_fx[free_index] = m_corners[i].x;
		m_cuda_fy[free_index] = m_corners[i].y;

		++free_index;
	}
	
	memset((void*)&m_cuda_index_map[n_new], 0, (MAX_NUM_FEATURES - n_new)*sizeof(m_cuda_index_map[0]));

	Caller_Tracker_Registration(m_cuda_fx, m_cuda_fy, n_new, m_cuda_index_map);

	
	// unlock the GPU here.
	m_locked = false; // yes, this can race too.		
	litmus_unlock(m_kexcluLock);		
	
	// Echo
	if (verbose) {
		cout << "CFeature2DPool::Select_GPU() : frame = " << frame 
			<< ", f_num = " << f_num << ", new_num = " << n_new << ", total_num = " << m_mapActiveID[frame].size() << endl;
	}

	m_stat.frame = frame;
	m_stat.time_select = 0;
	m_stat.n_new = n_new;
	m_mapStat[frame] = m_stat;

	// Update activeID map so that newly selected features are included.
	m_mapActiveID[frame] = m_listActiveID;

	return n_new;
}


//double RUNTIMES[500] = {0};
//int RTCOUNT = 0;

void CFeature2DPool::Track_GPU(int frame, IplImage* img, double* Hmg)
{	
	//static CTimer timer;
	
	//timer.Start();
	//double t0, t1, t2, t3;

	static bool first = true;
	
	m_curGPU = litmus_lock(m_kexcluLock);
	//printf("got GPU %d\n", m_curGPU);
	m_locked = true;	


	if(first)
	{
		set_cur_gpu(m_curGPU);
		first = false;
	}
	else if(!MigrateToGPU_Implicit(m_curGPU))
	{
		printf("MIGRATION FAILED!\n");
		litmus_unlock(m_kexcluLock);
		exit(-1);
	}
	
	/*
	int cudaDev = -1;
	cudaGetDevice(&cudaDev);
	printf("Using device %d (%d assigned).\n", cudaDev, m_curGPU);
	*/	
	
#if 1
	// TODO: Move this to a function call in cuda_image_align.cu
	// stage the data in GPU's linear memory.
	CUDA_SAFE_CALL( chunkMemcpy2D(Image2Temp[CUR_DEVICE][0], Image2TempPitch[CUR_DEVICE][0],
								  img->imageData, img->width*sizeof(float),
								  img->width*sizeof(float), img->height,
								  cudaMemcpyHostToDevice,
								  CUR_DEVICE,
								  useEngineLocks()));
	
	// Copy the data to array memory.  Local copies on the GPU seem to only
	// invoke the copy engine, not any kernel.  We don't to chunking here.
	// However, the copy should (?) be fast.
	if(useEngineLocks()) litmus_lock(CE_SEND_LOCKS[CUR_DEVICE]);
	CUDA_SAFE_CALL( cudaMemcpy2DToArrayAsync(Image2[CUR_DEVICE][0], 0, 0, 
							Image2Temp[CUR_DEVICE][0], Image2TempPitch[CUR_DEVICE][0],
							img->width*sizeof(float), img->height,
							cudaMemcpyDeviceToDevice, STREAMS[CUR_DEVICE]));
	CUDA_SAFE_CALL( cudaStreamSynchronize(STREAMS[CUR_DEVICE]));
	if(useEngineLocks()) litmus_unlock(CE_SEND_LOCKS[CUR_DEVICE]);
#else
	CUDA_SAFE_CALL( cudaMemcpy2DToArrayAsync(Image2[CUR_DEVICE][0], 0, 0, 
											 img->imageData, img->width*sizeof(float),
											 img->width*sizeof(float), img->height,
											 cudaMemcpyHostToDevice, STREAMS[CUR_DEVICE]));	
#endif
	
	
	float vec_invH[9] = {1, 0, 0, 0, 1, 0, 0, 0, 1};
	if (Hmg != NULL) {
		for (int i=0; i < 9; i++) vec_invH[i] = (float)Hmg[i];
	}

	//t0 = timer.Stop();
	Caller_Align(Image2[CUR_DEVICE][0], m_cuda_motion, m_cuda_residual, (Hmg != NULL) ? vec_invH : NULL);
	//t1 = timer.Stop();
	

	int n_total = (int)m_listActiveID.size();
	int n_success = 0;

	float failthres = 400.0f;
	int boundary = 13;
	int count_fail_alignment=0;
	int count_fail_bound=0;

	for (int i=0; i < MAX_NUM_FEATURES; i++)
	{
		int id = m_cuda_index_map_global[i];
		if (id == -1) continue;

		CFeature2DTrack* pTracker = GetFeatureTrackPtr(id);

		int status = 0;
		float rms = 0.0f, ncc = 0.0f, shrink = 0.0f, bias = 0.0f;


		float px[2], nx[2] = {0.0f, 0.0f};
		px[0] = m_cuda_motion[0*MAX_NUM_FEATURES+i] + m_cuda_fx[i];
		px[1] = m_cuda_motion[1*MAX_NUM_FEATURES+i] + m_cuda_fy[i];

		if (m_pCameraModel != NULL)
			m_pCameraModel->NormalizeFromTable(px[0], px[1], nx[0], nx[1]);

		if (m_cuda_residual[i] > failthres) {
			status = CImageAlign::TRACK_FAIL_ALIGNMENT;
		}
		else if ((px[0] < boundary) || (px[0] > img->width - boundary) ||
				 (px[1] < boundary) || (px[1] > img->height- boundary)) {
			status = CImageAlign::TRACK_FAIL_BOUND;
		}
		else {
			status = CImageAlign::TRACK_SUCCESS;
		}
		
		pTracker->push_back(new CFeature2D(px[0], px[1], frame, status, rms, ncc, shrink, bias, nx[0], nx[1]));
		
		if (status & CImageAlign::TRACK_FAIL) {
			m_listLostID.push_back(id);
			m_listActiveID.remove(id);
			m_cuda_index_map_lost[i] = 1;
			m_cuda_index_map_global[i] = -1;
		}
		else {
			pTracker->SetCurrentFrame(frame);
			pTracker->SetEndFrame(frame);
			n_success++;
		}
	}
	
	//t2 = timer.Stop();

	m_stat.time_select = 0;
	m_stat.n_new = 0;

	m_stat.frame = frame;
	m_stat.n_success = n_success;
	m_stat.n_total = n_total;
	m_stat.time_track = 0;
	
	//cout << "fail_bound = " << count_fail_bound ;
	//cout << "    count_fail_alignment = " << count_fail_alignment << endl;

	m_mapStat[frame] = m_stat;
	m_mapActiveID[frame] = m_listActiveID;
	
	//t3 = timer.Stop();
	
	//if(RTCOUNT < 500)
	//	RUNTIMES[RTCOUNT++] = 1e3*(t1 - t0);
	
	//printf("Kernel = %f, Loop = %f, Map = %f\n", 1e3*(t1 - t0), 1e3*(t2 - t1), 1e3*(t3 - t2));
}
